#include "../std.h"
#include "../jobs/job.h"
#include "../jobs/states.h"
#include "../jobs/users.h"
#include "../run/run.h"
#include "../run/run_plugin.h"

#include "plugins.h"

/*
  Substitution:
   %I - job id
*/


ContinuationPlugins::ContinuationPlugins(void) {
}

ContinuationPlugins::~ContinuationPlugins(void) {
}

bool ContinuationPlugins::add(job_state_t state,unsigned int timeout,const char* command) {
  if((state == JOB_STATE_ACCEPTED) ||
     (state == JOB_STATE_PREPARING) ||
     (state == JOB_STATE_SUBMITING) ||
     (state == JOB_STATE_FINISHING) ||
     (state == JOB_STATE_FINISHED) ||
     (state == JOB_STATE_DELETED)) {
    commands[state].cmd=command;
    commands[state].to=timeout;
  } else { return false; };
  return true;
}

bool ContinuationPlugins::add(const char* state,unsigned int timeout,const char* command) {
  job_state_t i = JobDescription::get_state(state);
  if(i != JOB_STATE_UNDEFINED) {
    return add(i,timeout,command);
  };
  return false;
}

bool ContinuationPlugins::run(const JobDescription &job,const JobUser& user,int& result,string& response) {
  job_state_t state = job.get_state();
  response.resize(0);
  if(commands[state].cmd.length() == 0) {
    result=0; return true;
  };
  string cmd = commands[state].cmd;
  for(int p = 0;;) {
    p=cmd.find('%',p);
    if(p==string::npos) break;
    if(cmd[p+1]=='I') {
      cmd.replace(p,2,job.get_id().c_str());
      p+=job.get_id().length();
    } else if(cmd[p+1]=='S') {
      cmd.replace(p,2,job.get_state_name());
      p+=strlen(job.get_state_name());
    } else {
      p+=2;
    };
  };
  if(!user.substitute(cmd)) {
    return false;
  };
  string res_out("");
  string res_err("");
  char** args = string_to_args(cmd);
  if(args == NULL) {
    return false;
  };
  if(!Run::plain_run_piped(args,NULL,&res_out,&res_err,commands[state].to,&result)) {
    free(args);
    return false;
  };
  response=res_out;
  if(res_err.length()) {
    if(response.length()) response+=" : ";
    response+=res_err;
  };
  free_args(args);
  return true;
}
